#include "compat_wrappers.h"
#include "extension.h"
#include "effect_dispatch_data.h"

inline void UTIL_DecalTrace( trace_t *pTrace, const char*decalName )
{
	static void *func = NULL;
	if(!func)
	{
		if(!g_pGameConf->GetMemSig("UTIL_DecalTrace", &func))
		{
			assert(0);
			return;
		}
	}
	typedef void (*_func)(trace_t *,const char*);
	_func thisfunc = (_func)(func);
	thisfunc(pTrace,decalName);
}

inline trace_t &GetTouchTrace( CBaseEntity *pEntity )
{
	static void *func = NULL;
	if(!func)
	{
		if(!g_pGameConf->GetMemSig("GetTouchTrace", &func))
		{
			assert(0);
			//trace_t ptr;
			//return ptr;
		}
	}
	typedef trace_t &(*_func)(void *);
	_func thisfunc = (_func)(func);
	return thisfunc(pEntity);
}

inline void PrecacheParticleSystem( const char *pName )
{
	static void *func = NULL;
	if(!func)
	{
		if(!g_pGameConf->GetMemSig("PrecacheParticleSystem", &func))
		{
			assert(0);
			return;
		}
	}
	typedef void (*_func)(const char *);
	_func thisfunc = (_func)(func);
	thisfunc(pName);
}

inline void DispatchEffect( const char *pName, CEffectData &data )
{
	static void *func = NULL;
	if(!func)
	{
		if(!g_pGameConf->GetMemSig("DispatchEffect", &func))
		{
			assert(0);
			return;
		}
	}
	typedef void (*_func)(const char *,CEffectData &);
	_func thisfunc = (_func)(func);
	thisfunc(pName,data);
}

inline void DispatchParticleEffect( int iEffectIndex, Vector vecOrigin, Vector vecStart, Vector vecAngles, CBaseEntity *pEntity )
{
	CEffectData	data;

	data.m_nHitBox = iEffectIndex;
	data.m_vOrigin = vecOrigin;
	data.m_vStart = vecStart;
	//data.m_vAngles = vecAngles;

	if ( pEntity )
	{
		int iEntIndex;
		edict_t *pEdict = gameents->BaseEntityToEdict(pEntity);
		if(pEdict) {
			iEntIndex = IndexOfEdict(pEdict);
		}
		data.m_nEntIndex = iEntIndex;
		data.m_fFlags |= PARTICLE_DISPATCH_FROM_ENTITY;
		data.m_nDamageType = PATTACH_CUSTOMORIGIN;
	}
	else
	{
		data.m_nEntIndex = 0;
	}
	DispatchEffect( "ParticleEffect", data );
}

inline int GetParticleSystemIndex( const char *pParticleSystemName )
{
	static void *func = NULL;
	if(!func)
	{
		if(!g_pGameConf->GetMemSig("GetParticleSystemIndex", &func))
		{
			assert(0);
			return 0;
		}
	}
	typedef int (*_func)(const char *);
	_func thisfunc = (_func)(func);
	return thisfunc(pParticleSystemName);
}

inline void SetParent(CBaseEntity *pEntity, CBaseEntity* pNewParent, int iAttachment = -1 )
{
	static void *func = NULL;
	if(!func)
	{
		if(!g_pGameConf->GetMemSig("SetParent", &func))
		{
			assert(0);
			return;
		}
	}
	typedef void (*_func)(void *,CBaseEntity *,int);
	_func thisfunc = (_func)(func);
	thisfunc(pEntity,pNewParent,iAttachment);
}

inline void UTIL_Remove(CBaseEntity *pEntity)
{
	static void *func = NULL;
	if(!func)
	{
		if(!g_pGameConf->GetMemSig("UTIL_Remove", &func))
		{
			assert(0);
			return;
		}
	}
	typedef void (*_func)(CBaseEntity *);
	_func thisfunc = (_func)(func);
	thisfunc(pEntity);
}

inline int HasCondition(CBaseEntity *pEntity, int num)
{
	static void *func = NULL;
	if(!func)
	{
		if(!g_pGameConf->GetMemSig("HasCondition", &func))
		{
			assert(0);
			return 0;
		}
	}
	typedef bool (*_func)(void *,int);
	_func thisfunc = (_func)(func);
	return thisfunc(pEntity,num);
}

inline void InitDefaultAIRelationships()
{
	static void *func = NULL;
	if(!func)
	{
		if(!g_pGameConf->GetMemSig("InitDefaultAIRelationships", &func))
		{
			assert(0);
			return;
		}
	}
	typedef void (*_func)();
	_func thisfunc = (_func)(func);
	thisfunc();
}

inline Vector Weapon_ShootPosition(CBaseEntity *pEntity)
{
	static void *func = NULL;
	if(!func)
	{
		if(!g_pGameConf->GetMemSig("Weapon_ShootPosition", &func))
		{
			assert(0);
		}
	}
	typedef Vector (*_func)(void*);
	_func thisfunc = (_func)(func);
	return thisfunc(pEntity);
}

inline Vector GetActualShootTrajectory(CBaseEntity *pEntity, Vector &shootOrigin)
{
	static void *func = NULL;
	if(!func)
	{
		if(!g_pGameConf->GetMemSig("GetActualShootTrajectory", &func))
		{
			assert(0);
		}
	}
	typedef Vector (*_func)(void*,Vector&);
	_func thisfunc = (_func)(func);
	return thisfunc(pEntity,shootOrigin);
}

inline void DoImpactEffect(CBaseEntity *pEntity, trace_t &tr, int nDamageType)
{
	static void *func = NULL;
	if(!func)
	{
		if(!g_pGameConf->GetMemSig("DoImpactEffect", &func))
		{
			assert(0);
			return;
		}
	}
	typedef void (*_func)(void *, trace_t &, int);
	_func thisfunc = (_func)(func);
	thisfunc(pEntity, tr, nDamageType);
}

inline CBaseEntity *CreateCombineBall(Vector &origin, Vector &velocity, float radius, float mass, float lifetime, CBaseEntity *pOwner)
{
	static void *func = NULL;
	if(!func)
	{
		if(!g_pGameConf->GetMemSig("CreateCombineBall", &func))
		{
			assert(0);
			return NULL;
		}
	}
	typedef CBaseEntity *(*_func)(Vector &, Vector &, float, float, float, CBaseEntity *);
	_func thisfunc = (_func)(func);
	return thisfunc(origin, velocity, radius, mass, lifetime, pOwner);
}

inline Activity ActivityList_IndexForName(const char *pszActivityName)
{
	static void *func = NULL;
	if(!func)
	{
		if(!g_pGameConf->GetMemSig("ActivityList_IndexForName", &func))
		{
			assert(0);
			return (Activity)0;
		}
	}
	typedef Activity (*_func)(const char *);
	_func thisfunc = (_func)(func);
	return thisfunc(pszActivityName);
}

inline void UTIL_Tracer(Vector &vecStart, Vector &vecEnd, int iEntIndex, int iAttachment, float flVelocity, bool bWhiz, const char *pCustomTracerName, int iParticleID)
{
	static void *func = NULL;
	if(!func)
	{
		if(!g_pGameConf->GetMemSig("UTIL_Tracer", &func))
		{
			assert(0);
			return;
		}
	}
	typedef void (*_func)(Vector &, Vector &, int, int, float, bool, const char *, int);
	_func thisfunc = (_func)(func);
	thisfunc(vecStart, vecEnd, iEntIndex, iAttachment, flVelocity, bWhiz, pCustomTracerName, iParticleID);
}

inline void UTIL_TraceLine(const Vector &vecAbsStart, const Vector &vecAbsEnd, unsigned int mask, CBaseEntity *ignore, int collisionGroup, trace_t *ptr)
{
	static void *func = NULL;
	if(!func)
	{
		if(!g_pGameConf->GetMemSig("UTIL_TraceLine", &func))
		{
			assert(0);
			return;
		}
	}
	typedef void (*_func)(const Vector &, const Vector &, unsigned int, CBaseEntity *, int, trace_t *);
	_func thisfunc = (_func)(func);
	thisfunc(vecAbsStart,vecAbsEnd,mask,ignore,collisionGroup,ptr);
}

inline int EventList_IndexForName(const char *pszEventName)
{
	static void *func = NULL;
	if(!func)
	{
		if(!g_pGameConf->GetMemSig("EventList_IndexForName", &func))
		{
			assert(0);
			return 0;
		}
	}
	typedef int (*_func)(const char *);
	_func thisfunc = (_func)(func);
	return thisfunc(pszEventName);
}

inline const char *EventList_NameForIndex( int iEventIndex )
{
	static void *func = NULL;
	if(!func)
	{
		if(!g_pGameConf->GetMemSig("EventList_NameForIndex", &func))
		{
			assert(0);
			return NULL;
		}
	}
	typedef const char *(*_func)(int);
	_func thisfunc = (_func)(func);
	return thisfunc(iEventIndex);
}

inline void UTIL_PredictedPosition( CBaseEntity *pTarget, float flTimeDelta, Vector *vecPredictedPosition )
{
	static void *func = NULL;
	if(!func)
	{
		if(!g_pGameConf->GetMemSig("UTIL_PredictedPosition", &func))
		{
			assert(0);
			return;
		}
	}
	typedef void (*_func)(CBaseEntity *,float,Vector *);
	_func thisfunc = (_func)(func);
	thisfunc(pTarget,flTimeDelta,vecPredictedPosition);
}

inline Vector BodyTarget( CBaseEntity *entity, const Vector &posSrc, bool bNoisy = true )
{
	static void *func = NULL;
	if(!func)
	{
		if(!g_pGameConf->GetMemSig("BodyTarget", &func))
		{
			assert(0);
		}
	}
	typedef Vector (*_func)(void*, const Vector &,bool);
	_func thisfunc = (_func)(func);
	return thisfunc(entity,posSrc,bNoisy);
}

inline Vector GetAbsOrigin(CBaseEntity *pEntity)
{
	edict_t *pEdict = gameents->BaseEntityToEdict(pEntity);
	if (!pEdict || pEdict->IsFree())
		return vec3_origin;
	return (Vector&)pEdict->GetNetworkable()->GetPVSInfo()->m_vCenter;
}

inline CBaseEntity *UTIL_PlayerByIndex(int num)
{
	edict_t *pEdict = PEntityOfEntIndex(num);
	if (!pEdict || pEdict->IsFree())
		return NULL;
	if (num > 0 && num <= playerhelpers->GetMaxClients()) {
		IGamePlayer *pPlayer = playerhelpers->GetGamePlayer(pEdict);
		if (!pPlayer || !pPlayer->IsConnected()) {
			return NULL;
		}
	}
	else {
		return NULL;
	}
	IServerUnknown *pUnk;
	if ((pUnk=pEdict->GetUnknown()) == NULL)
		return NULL;
	return pUnk->GetBaseEntity();
}

inline int GetEntProp(CBaseEntity *pEntity, const char *str, bool IsBool = false, bool send = false) 
{
	datamap_t *pMap = gamehelpers->GetDataMap(pEntity);
	if(pMap == NULL)
		return 0;
	typedescription_t *pType = gamehelpers->FindInDataMap(pMap,str);
	if(pType == NULL)
		return 0;
	int offset = GetTypeDescOffs(pType);
	if(offset == -1)
		return 0;
	if(send) {
		sm_sendprop_info_t info;
		IServerUnknown *pUnk = (IServerUnknown *)pEntity;
		IServerNetworkable *pNet = pUnk->GetNetworkable();
		if(!pNet)
			return 0;
		if(!gamehelpers->FindSendPropInfo(pNet->GetServerClass()->GetName(), str, &info))
			return 0;
		offset = info.actual_offset;
	}
	if(IsBool) {
		return *(bool *)((uint8_t *)pEntity + offset) ? 1 : 0;
	}
	return *(int *)((uint8_t *)pEntity + offset);
}

inline void SetEntProp(CBaseEntity *pEntity, const char *str, int value, bool IsBool = false, bool send = false) 
{
	datamap_t *pMap = gamehelpers->GetDataMap(pEntity);
	if(pMap == NULL)
		return;
	typedescription_t *pType = gamehelpers->FindInDataMap(pMap,str);
	if(pType == NULL)
		return;
	int offset = GetTypeDescOffs(pType);
	if(offset == -1)
		return;
	if(send) {
		sm_sendprop_info_t info;
		IServerUnknown *pUnk = (IServerUnknown *)pEntity;
		IServerNetworkable *pNet = pUnk->GetNetworkable();
		if(!pNet)
			return;
		if(!gamehelpers->FindSendPropInfo(pNet->GetServerClass()->GetName(), str, &info))
			return;
		offset = info.actual_offset;
	}
	if(IsBool) {
		*(bool *)((uint8_t *)pEntity + offset) = value ? true : false;
		return;
	}
	*(int *)((uint8_t *)pEntity + offset) = value;
}

inline void SetEntPropFloat(CBaseEntity *pEntity, const char *str, float value, bool send = false) 
{
	datamap_t *pMap = gamehelpers->GetDataMap(pEntity);
	if(pMap == NULL)
		return;
	typedescription_t *pType = gamehelpers->FindInDataMap(pMap,str);
	if(pType == NULL)
		return;
	int offset = GetTypeDescOffs(pType);
	if(offset == -1)
		return;
	if(send) {
		sm_sendprop_info_t info;
		IServerUnknown *pUnk = (IServerUnknown *)pEntity;
		IServerNetworkable *pNet = pUnk->GetNetworkable();
		if(!pNet)
			return;
		if(!gamehelpers->FindSendPropInfo(pNet->GetServerClass()->GetName(), str, &info))
			return;
		offset = info.actual_offset;
	}
	*(float *)((uint8_t *)pEntity + offset) = value;
}

inline float GetEntPropFloat(CBaseEntity *pEntity, const char *str, bool IsBool = false, bool send = false) 
{
	datamap_t *pMap = gamehelpers->GetDataMap(pEntity);
	if(pMap == NULL)
		return 0.0;
	typedescription_t *pType = gamehelpers->FindInDataMap(pMap,str);
	if(pType == NULL)
		return 0.0;
	int offset = GetTypeDescOffs(pType);
	if(offset == -1)
		return 0.0;
	if(send) {
		sm_sendprop_info_t info;
		IServerUnknown *pUnk = (IServerUnknown *)pEntity;
		IServerNetworkable *pNet = pUnk->GetNetworkable();
		if(!pNet)
			return 0.0;
		if(!gamehelpers->FindSendPropInfo(pNet->GetServerClass()->GetName(), str, &info))
			return 0.0;
		offset = info.actual_offset;
	}
	if(IsBool) {
		return *(bool *)((uint8_t *)pEntity + offset) ? 1 : 0;
	}
	return *(float *)((uint8_t *)pEntity + offset);
}

inline Vector GetEntPropVector(CBaseEntity *pEntity, const char *str) 
{
	datamap_t *pMap = gamehelpers->GetDataMap(pEntity);
	if(pMap == NULL)
		return vec3_origin;
	typedescription_t *pType = gamehelpers->FindInDataMap(pMap,str);
	if(pType == NULL)
		return vec3_origin;
	int offset = GetTypeDescOffs(pType);
	if(offset == -1)
		return vec3_origin;
	return *(Vector *)((uint8_t *)pEntity + offset);
}


inline void SetEntPropVector(CBaseEntity *pEntity, const char *str, Vector m_vec) 
{
	datamap_t *pMap = gamehelpers->GetDataMap(pEntity);
	if(pMap == NULL)
		return;
	typedescription_t *pType = gamehelpers->FindInDataMap(pMap,str);
	if(pType == NULL)
		return;
	int offset = GetTypeDescOffs(pType);
	if(offset == -1)
		return;
	*(Vector *)((uint8_t *)pEntity + offset) = m_vec;
}

inline CBaseEntity *GetEntPropEnt(CBaseEntity *pEntity, const char *str) 
{
	datamap_t *pMap = gamehelpers->GetDataMap(pEntity);
	if(pMap == NULL)
		return NULL;
	typedescription_t *pType = gamehelpers->FindInDataMap(pMap,str);
	if(pType == NULL)
		return NULL;
	int offset = GetTypeDescOffs(pType);
	if(offset == -1)
		return NULL;
	CBaseHandle &hndl = *(CBaseHandle *)((uint8_t *)pEntity + offset);
	return gameents->EdictToBaseEntity(PEntityOfEntIndex(hndl.GetEntryIndex()));
}

inline void SetEntPropEnt(CBaseEntity *pEntity, const char *str, CBaseEntity *entity) 
{
	datamap_t *pMap = gamehelpers->GetDataMap(pEntity);
	if(pMap == NULL)
		return;
	typedescription_t *pType = gamehelpers->FindInDataMap(pMap,str);
	if(pType == NULL)
		return;
	int offset = GetTypeDescOffs(pType);
	if(offset == -1)
		return;
	CBaseHandle &hndl = *(CBaseHandle *)((uint8_t *)pEntity + offset);
	if(entity != NULL) {
		IHandleEntity *pHandleEnt = (IHandleEntity *)entity;
		hndl.Set(pHandleEnt);
	}
	else {
		hndl.Set(NULL);
	}
}

inline const char *GetEntPropString(CBaseEntity *pEntity, const char *str) 
{
	datamap_t *pMap = gamehelpers->GetDataMap(pEntity);
	if(pMap == NULL)
		return NULL;
	typedescription_t *pType = gamehelpers->FindInDataMap(pMap,str);
	if(pType == NULL)
		return NULL;
	int offset = GetTypeDescOffs(pType);
	if(offset == -1)
		return NULL;
	string_t s;
	if ((s = *(string_t *)((uint8_t *)pEntity + offset)) == NULL_STRING)
		return NULL;
	return STRING(s);
}

inline bool FClassnameIs(CBaseEntity *pEntity, const char *szClassname)
{
	edict_t *pEdict = gameents->BaseEntityToEdict(pEntity);
	if (!pEdict || pEdict->IsFree())
		return false;
	return !strcmp( pEdict->GetClassName(), szClassname ) ? true : false; 
}

inline bool IsNpc(CBaseEntity *pEntity)
{
	edict_t *pEdict = gameents->BaseEntityToEdict(pEntity);
	if(!pEdict)
		return false;
	return !strncmp( pEdict->GetClassName(), "npc_", 4 ) ? true : false; 
}

inline bool IsWeapon(CBaseEntity *pEntity)
{
	edict_t *pEdict = gameents->BaseEntityToEdict(pEntity);
	if(!pEdict)
		return false;
	return !strncmp( pEdict->GetClassName(), "weapon_", 7 ) ? true : false; 
}

inline bool IsPlayer(CBaseEntity *pEntity)
{
	edict_t *pEdict = gameents->BaseEntityToEdict(pEntity);
	if(!pEdict) {
		return false;
	}
	int ent = IndexOfEdict(pEdict);
	if (ent > 0 && ent <= playerhelpers->GetMaxClients()) {
		return true;
	}
	return false;
}

inline bool IsFakeClient(CBaseEntity *pEntity) 
{
	edict_t *pEdict = gameents->BaseEntityToEdict(pEntity);
	if(!pEdict) {
		return false;
	}
	IGamePlayer *pPlayer = playerhelpers->GetGamePlayer(pEdict);
	if (!pPlayer || !pPlayer->IsConnected()) {
		return false;
	}
	if(pPlayer->IsFakeClient()) {
		return true;
	}
	return false;
}

inline Vector ApplySpread( Vector &vecSpread, Vector &vecDir, float bias = 1.0  )
{
	float x, y, z;
	if ( bias > 1.0 )
		bias = 1.0;
	else if ( bias < 0.0 )
		bias = 0.0;
	float shotBiasMin = -1.0;
	float shotBiasMax = 1.0;
	float shotBias = ( ( shotBiasMax - shotBiasMin ) * bias ) + shotBiasMin;
	float flatness = ( fabsf(shotBias) * 0.5 );
	do
	{
		x = RandomFloat(-1,1) * flatness + RandomFloat(-1,1) * (1 - flatness);
		y = RandomFloat(-1,1) * flatness + RandomFloat(-1,1) * (1 - flatness);
		if ( shotBias < 0 )
		{
			x = ( x >= 0 ) ? 1.0 - x : -1.0 - x;
			y = ( y >= 0 ) ? 1.0 - y : -1.0 - y;
		}
		z = x*x+y*y;
	} while (z > 1);
	Vector m_vecRight;
	Vector m_vecUp;
	VectorVectors( vecDir, m_vecRight, m_vecUp );
	return vecDir + x * vecSpread.x * m_vecRight + y * vecSpread.y * m_vecUp;
}

inline bool HasSpawnFlags(CBaseEntity *entity, int nFlags )
{
	if(entity == NULL) {
		return false;
	}
	int m_spawnflags = GetEntProp(entity,"m_spawnflags");
	return (m_spawnflags & nFlags) != 0; 
}

inline void AddSpawnFlags(CBaseEntity *entity, int nFlags ) 
{
	int m_spawnflags = GetEntProp(entity,"m_spawnflags");
	m_spawnflags |= nFlags;
	SetEntProp(entity,"m_spawnflags",m_spawnflags); 
}

inline void RemoveSpawnFlags(CBaseEntity *entity, int nFlags ) 
{
	int m_spawnflags = GetEntProp(entity,"m_spawnflags");
	m_spawnflags &= ~nFlags;
	SetEntProp(entity,"m_spawnflags",m_spawnflags); 
}

inline float UTIL_DistApprox( const Vector &vec1, const Vector &vec2 )
{
	float dx;
	float dy;
	float dz;
	dx = vec1.x - vec2.x;
	dy = vec1.y - vec2.y;
	dz = vec1.z - vec2.z;
	return fabs(dx) + fabs(dy) + fabs(dz);
}

inline Vector VecCheckThrowTolerance( CBaseEntity *pEdict, const Vector &vecSpot1, Vector vecSpot2, float flSpeed, float flTolerance )
{
	flSpeed = MAX( 1.0f, flSpeed );
	float flGravity = 600.0;
	Vector vecGrenadeVel = (vecSpot2 - vecSpot1);
	float time = vecGrenadeVel.Length( ) / flSpeed;
	vecGrenadeVel = vecGrenadeVel * (1.0 / time);
	vecGrenadeVel.z += flGravity * time * 0.5;
	Vector vecApex = vecSpot1 + (vecSpot2 - vecSpot1) * 0.5;
	vecApex.z += 0.5 * flGravity * (time * 0.5) * (time * 0.5);
	trace_t tr;
	UTIL_TraceLine( vecSpot1, vecApex, MASK_SOLID, pEdict, COLLISION_GROUP_NONE, &tr );
	if (tr.fraction != 1.0)
	{
		return vec3_origin;
	}
	UTIL_TraceLine( vecApex, vecSpot2, MASK_SOLID_BRUSHONLY, pEdict, COLLISION_GROUP_NONE, &tr );
	if ( tr.fraction != 1.0 )
	{
		bool bFail = true;
		if ( flTolerance > 0.0f )
		{
			float flNearness = ( tr.endpos - vecSpot2 ).LengthSqr();
			if ( flNearness < Square( flTolerance ) )
			{
				bFail = false;
			}
		}
		if ( bFail )
		{
			return vec3_origin;
		}
	}
	return vecGrenadeVel;
}

inline bool IsSolidFlagSet( CBaseEntity *entity, int flagMask )
{
	int m_usSolidFlags = GetEntProp(entity,"m_usSolidFlags",false,true);
	return (m_usSolidFlags & flagMask) != 0;
}
